from simtek_graphics import *
from pylab import *

"""
Simplified particle simulator:

Models particle-spring ODE systems. Solves the ODE by time-stepping
with forward Euler or Trapezoid with simple fixed-point
iteration. Simplified version of particle_simulator.py.

Choose initial data by setting values of "x" and "v" below.

The solution vector u has 4 blocks [x1 x2 v1 v2] representing two
spatial components of the position x and two spatial components of the
velocity v. Each block is then the size of the number of particles.

The model f also has 4 blocks [v1 v2 y1 y2] representing two spatial
components of the velocity v and two spatial components of the
acceleration (force divided by mass) y. Each block is then the size of
the number of particles.

To make the model easier to read/implement, the blocks are reformatted
(unpacked) into two-dimensional arrays of shape (M, 2), where M is the
number of particles. For example, the "x" component of the position of
particle i is x[i, 0] and the Euclidean norm of the velocity for
particle i can be computed as: norm(v[i, :]).

The model is implemented in the function f3body(t, u), where the force
is computed for all pairs of particles. An elastic spring force is
implemented, but it's possible to modify the force to instead model a gravity
force for example.

"""

M = 3                  # Number of particles

u = zeros(4*M)         # 1-dimensional array of size 
                       # number of equations/particle (x and y)
                       # times number of particles
u0 = zeros(4*M)
fval = zeros(4*M)
x = zeros([M, 2])      # Array of M number of rows and two 
                       # columns (for x(t) and y(t))

v = zeros([M, 2])      # Array of M number of rows and two 
                       # columns (for v_x(t) and v_y(t))

a = zeros([M, 2])      # Vector to collect the acceleration F_x/m and F_y/m


radius = zeros(M)      
radius += 0.5

# Initial values
x[0, 0] = 0.0          # Particle 1 starts in x=0
x[0, 1] = 0.0          # Particle 1 starts in y=0
x[1, 0] = 1.0          # Particle 2 starts in x=1
x[1, 1] = 1.4          # Particle 2 starts in y=1.4
x[2, 0] = 2.0          # Particle 3 starts in x=2
x[2, 1] = 0.0          # Particle 3 starts in y=0



v[0, 0] = -1.0         # Particle 1 has velocity -1 in x-dir
v[0, 1] =  0.0         # Particle 1 has velocity 0 in y-dir
v[1, 0] = 0.0          # Particle 2 has velocity 0 in x-dir
v[1, 1] = 1.0          # Particle 2 has velocity 1 in y-dir
v[2, 0] = 1.0          # Particle 3 has velocity 1 in x-dir
v[2, 1] = 1.0          # Particle 3 has velocity 1 in y-dir

# Pack values into u   
u[0*M:1*M] = x[:, 0]
u[1*M:2*M] = x[:, 1]
u[2*M:3*M] = v[:, 0]
u[3*M:4*M] = v[:, 1]


def f3body(t, u):
  # Unpack values into x and v
  x[:, 0] = u[0*M:1*M]
  x[:, 1] = u[1*M:2*M]
  v[:, 0] = u[2*M:3*M]
  v[:, 1] = u[3*M:4*M]

  a = zeros((M, 2))

  m = 1.0    # Mass
  E = 100.0  # Stiffness coefficient
  B = 4.0    # Damping coefficient
  L = 2.0    # Spring rest length

  for i0 in range(0, M):
      for i1 in range(0, M):
          # Don't compute force with self
          if(i0 == i1):
              continue

          r = norm(x[i1, :] - x[i0, :])
          e = (x[i1, :] - x[i0, :]) / r
          vr = v[i1, :] - v[i0, :]
          
          # Elastic spring force
          F = E*(r - L)*e

          # Damping spring force
          D = B*dot(vr, e)*e

          # Gravity force
          G = 0 # Add this yourself

          a[i0, :] += (1.0*F + 1.0*D + 0.0*G) / m

  # Pack values into fval
  fval = zeros(4*M)
  fval[0*M:1*M] = v[:, 0]     # Right hand side for x'(t), y'(t) (velocities)
  fval[1*M:2*M] = v[:, 1]
  fval[2*M:3*M] = a[:, 0]     # Right hand side for v_x'(t) and v_y'(t) (forces/mass)
  fval[3*M:4*M] = a[:, 1]

  return fval

f = f3body

init_graphics(x, radius, M)

dt = 0.01
t = 0.0
T = 10.0

counter = 0
while(game.running and t < T):

  # Time-stepping
  u0[:] = u[:]
  t0=t
  t=t+dt
  for i in range(0, 5):
      fval = f(t, u)
      fval0 = f(t0, u0)
      u = u0 + 0.5 * dt * (fval + fval0)

  xs = array([[0.0, u[0]]])
  vs = array([[0.0, u[1]]])

  
  
  # Unpack values into x, v and a (time derivative of v)
  x[:, 0] = u[0*M:1*M]
  x[:, 1] = u[1*M:2*M]
  v[:, 0] = u[2*M:3*M]
  v[:, 1] = u[3*M:4*M]
  a[:, 0] = fval[2*M:3*M]
  a[:, 1] = fval[3*M:4*M]

  update_graphics(counter, x, dt)

  counter += 1


