import os, sys, math, time

from scipy import weave
from scipy.weave import converters
from simtek_graphics import *
from spring_util import *
from pylab import *
from wave_data import *
from elastic_data import *

"""
Particle simulator:

Models particle-spring ODE systems representing an elastic string
(wave propagation) or solid elastic bodies with contact. Solves the
ODE by time-stepping with forward Euler or Trapezoid with simple
fixed-point iteration.

Choose initial data by commenting/uncommenting wave_data() or
elastic_data() just below.

The solution vector u has 4 blocks [x1 x2 v1 v2] representing two
spatial components of the position x and two spatial components of the
velocity v. Each block is then the size of the number of particles.

The model f also has 4 blocks [v1 v2 y1 y2] representing two spatial
components of the velocity v and two spatial components of the
acceleration (force divided by mass) y. Each block is then the size of
the number of particles.

To make the model easier to read/implement, the blocks are reformatted
(unpacked) into two-dimensional arrays of shape (M, 2), where M is the
number of particles. For example, the "x" component of the position of
particle i is x[i, 0] and the Euclidean norm of the velocity for
particle i can be computed as: norm(v[i, :]).

There are two implementations of the model f:

fsimple - pure Python implementation, easier to read but a bit
inefficient due to use of loops.

finline - the loops are written in C, which is inlined into the Python
code using SciPy.Weave (C code is compiled). Should be as efficient as
C, but the rest of the code is Python so easier to read/understand.

"""

(x, v, c, cmat, m, g, E, L, B, M, C, dt, radius, contact, method) = \
    wave_data()
#(x, v, c, cmat, m, g, E, L, B, M, C, dt, radius, contact, method) = \
#    elastic_data()

if(contact):
  import scikits.ann as ann

u = zeros(4*M)
u0 = zeros(4*M)
y = zeros([M, 2])

t = 0.0

# Pack values into u
u[0*M:1*M] = x[:, 0]
u[1*M:2*M] = x[:, 1]
u[2*M:3*M] = v[:, 0]
u[3*M:4*M] = v[:, 1]

def fsimple(t, u):

  # Unpack values into x and v
  x[:, 0] = u[0*M:1*M]
  x[:, 1] = u[1*M:2*M]
  v[:, 0] = u[2*M:3*M]
  v[:, 1] = u[3*M:4*M]

  y = zeros((M, 2))

  for i in range(0, C):
    i0 = c[i, 0]
    i1 = c[i, 1]

    r = norm(x[i1, :] - x[i0, :])
    e = (x[i1, :] - x[i0, :]) / r
    vr = v[i1, :] - v[i0, :]

    F = E*(r - L[i0, i1])*e
    G = B*dot(vr, e)*e

    y[i0, :] += (F + G) / m[i0]

  if(contact):
    # Compute contact forces
    for i in range(0, M):
      for j in likely[i]:
        if(i != j):
          p = x[i, :]
          r = norm(x[i, :] - x[j, :])
          if(cmat[j, i] == False and cmat[i, j] == False and \
               r < (radius[i] + radius[j])):
            e = (x[i, :] - x[j, :]) / r
            vr = v[i, :] - v[j, :]
            vpr = dot(vr, e)*e
            vpp = vr - vpr
            fr = 0.4
            F = E*(radius[i] + radius[j] - r)*e - B*vpr
            FF = -fr*norm(F)*vpp
            y[i, :] += (F + FF) / m[i]
            y[j, :] -= (F + FF) / m[j]


  y += g

  # Pack values into fval
  fval = zeros(4*M)
  fval[0*M:1*M] = v[:, 0]
  fval[1*M:2*M] = v[:, 1]
  fval[2*M:3*M] = y[:, 0]
  fval[3*M:4*M] = y[:, 1]

  return fval

def finline(t, u):

  # Unpack values into x and v
  x[:, 0] = u[0*M:1*M]
  x[:, 1] = u[1*M:2*M]
  v[:, 0] = u[2*M:3*M]
  v[:, 1] = u[3*M:4*M]

  y = zeros((M, 2))

  code = """
  double F, r, e, G;
  for(int i = 0; i < C; i++)
  {
    int i0 = c(i, 0);
    int i1 = c(i, 1);

    for(int j = 0; j < 2; j++)
    {
      r = sqrt(pow(x(i1, 0) - x(i0, 0), 2) + pow(x(i1, 1) - x(i0, 1), 2));
      e = (x(i1, j) - x(i0, j)) / r;

      double e0 = (x(i1, 0) - x(i0, 0)) / r;
      double e1 = (x(i1, 1) - x(i0, 1)) / r;

      F = E*(r - L(i0, i1))*e;

      double vr0 = v(i1, 0) - v(i0, 0);
      double vr1 = v(i1, 1) - v(i0, 1);

      G = B*(vr0*e0 + vr1*e1)*e;
      double Ftot = (F + G);

      y(i0, j) +=  Ftot / m(i0);
    }
  }
  """

  err = weave.inline(code,
                     ['E', 'B', 'L', 'x', 'v', 'y', 'm', 'g', 'M', 'c', 'C'],
                     type_converters=converters.blitz,
                     compiler = 'gcc')


  if(contact):
    # Compute contact forces
    for i in range(0, M):
      for j in likely[i]:
        if(i != j):
          p = x[i, :]
          r = norm(x[i, :] - x[j, :])
          if(cmat[j, i] == False and cmat[i, j] == False and \
               r < (radius[i] + radius[j])):
            e = (x[i, :] - x[j, :]) / r
            vr = v[i, :] - v[j, :]
            vpr = dot(vr, e)*e
            vpp = vr - vpr
            fr = 0.4
            F = E*(radius[i] + radius[j] - r)*e - B*vpr
            FF = -fr*norm(F)*vpp
            y[i, :] += 0.5* (F + FF) / m[i]
            y[j, :] -= 0.5 * (F + FF) / m[j]

  y += g
  # Pack values into fval
  fval = zeros(4*M)
  fval[0*M:1*M] = v[:, 0]
  fval[1*M:2*M] = v[:, 1]
  fval[2*M:3*M] = y[:, 0]
  fval[3*M:4*M] = y[:, 1]

  return fval

f = finline

init_graphics(x, radius, M)

start = time.clock()

N = 1000000
tarr = zeros(N); xarr = zeros(N); varr = zeros(N); farr = zeros(N)

T = 20

if(contact):
  tree = ann.kdtree(x)
else:
  tree = 0

counter = 0
while(game.running and t < T):

  if(contact):
    # Collision detection data structure
    tree = ann.kdtree(x)
    likely = tree.knn(x, 4)[0]

  # Time-stepping
  u0[:] = u[:]

  if(method == "Euler"):
    fval0 = f(t, u0)
    u = u0 + dt * fval0
  elif(method == "Midpoint"):
    for i in range(0, 5):
      fval = f(t, u)
      fval0 = f(t, u0)
      u = u0 + 0.5 * dt * (fval + fval0)

  t = t + dt
  
  # Unpack values into x, v and y (time derivative of v)
  x[:, 0] = u[0*M:1*M]
  x[:, 1] = u[1*M:2*M]
  v[:, 0] = u[2*M:3*M]
  v[:, 1] = u[3*M:4*M]
  y[:, 0] = fval0[2*M:3*M]
  y[:, 1] = fval0[3*M:4*M]

  # Store values
  tarr[counter] = t
  xarr[counter] = x[M/2][1]
  varr[counter] = v[M/2][1]
  farr[counter] = y[M/2][1]

  update_graphics(counter, x, dt)

  counter += 1

end = time.clock() - start
print "elapsed: %f" % end

#spring_plot(tarr, xarr, varr, farr, counter)
